MODULE CoarseMand;

(*
 *  This program is copyright (c) Raoul Gough 1985 - 1990
 *
 *  Permission is granted to QUT to distribute this program
 *  as a demonstration program with the Gardens Point Modula
 *  compiler system
 *
 *  This program may be copied and or modified by users 
 *  for non-commercial purposes provided that this copyright 
 *  notice remains in the text
 *
 *  All other uses require the written permission of the 
 *  copyright holder.
 *)

(* Revision record:

   30 Aug 90  John Hynd		Output xc,yc,mag instead of buttons
				Accept window location and size arguments

*)

FROM ProgArgs IMPORT ArgNumber, GetArg;
FROM IntStr IMPORT Value;
FROM InOut IMPORT WriteString, WriteLn, WriteInt, WriteCard;
FROM RealInOut IMPORT ReadReal, WriteReal;
FROM RealMath IMPORT round;
FROM Graphics IMPORT GetMouseClick, PlaceGraphWindow, CloseScreen, HiResDot, Flush,
			DrawBlock, Dot, EventHappened;

CONST
  LeftMouseButton = 256;
  RightMouseButton = 1024;

  NumColours = 60;
  Skip = 1;		(* Use only every Skip'th colour *)
  ColourOffset = 13;	(* Avoid window manager and primary colours *)

TYPE
  Complex = RECORD
              real, imag : REAL
            END;

VAR
  left, top, xMax, yMax : INTEGER;	(* Window parameters
					   - should be CARDINAL*)
  xc, yc, xmag, ymag, magnitudeLimit : REAL;
  squaresLimit : CARDINAL;


  PROCEDURE MaxSquares (z : Complex) : CARDINAL;

  VAR
    count : CARDINAL;
    rsquared, isquared : REAL;
    temp : Complex;
    over : BOOLEAN;

  BEGIN (* MaxSquares *)
    temp := z;
    count := 1;
    WITH temp DO
      REPEAT
	rsquared := real * real;
	isquared := imag * imag;
	over := isquared + rsquared >= magnitudeLimit;
	IF NOT over THEN
	  INC (count);
	  imag := 2.0 * real * imag + z.imag;
	  real := rsquared - isquared + z.real;
	END; (* IF *)
      UNTIL over OR (count = squaresLimit);
    END; (* WITH *)
    RETURN count;
  END MaxSquares;

  PROCEDURE DrawMandelbrot;

  VAR
    x, y : INTEGER;
    z : Complex;
    colour : CARDINAL;
    done : BOOLEAN;
    step : INTEGER;

  BEGIN (* DrawMandelbrot *)
    done := FALSE;
    (*
     *  Do first scan
     *)
    y := 0;
    WHILE (y < yMax) AND NOT done DO
      x := 0;
      WHILE (x < xMax) AND NOT done DO
        WITH z DO
	  real := FLOAT (x - xMax DIV 2) / xmag + xc;
	  imag := FLOAT (yMax DIV 2 - y) / ymag + yc;
	  colour := (MaxSquares (z) * Skip MOD NumColours) + ColourOffset;
	  DrawBlock (x, y, 16, 16, colour);
        END; (* WITH  *)
	INC (x,16);
      END; (* FOR *)
      done := EventHappened ();
      INC (y,16);
    END; (* FOR *)
    step := 8;
    (*
     *  Now do NxN blocks
     *)
    WHILE step > 1 DO
      y := 0;
      WHILE (y < yMax) AND NOT done DO
        x := 0;
        WHILE (x < xMax) AND NOT done DO
  	  IF ODD (x DIV step) OR ODD (y DIV step) THEN
            WITH z DO
	      real := FLOAT (x - xMax DIV 2) / xmag + xc;
	      imag := FLOAT (yMax DIV 2 - y) / ymag + yc;
	      colour := (MaxSquares (z) * Skip MOD NumColours) + ColourOffset;
	      DrawBlock (x, y, step, step, colour);
            END; (* WITH  *)
	  END;
	  INC (x,step);
        END; (* FOR *)
        done := EventHappened ();
        INC (y,step);
      END; (* FOR *)
      step := step DIV 2;
    END;
    (*
     *  Now do  lines in high res
     *)
    y := 0;
    WHILE (y < yMax) AND NOT done DO
      x := 0;
      WHILE (x < xMax) AND NOT done DO
	IF ODD (x) OR ODD (y) THEN
          WITH z DO
	    real := FLOAT (x - xMax DIV 2) / xmag + xc;
	    imag := FLOAT (yMax DIV 2 - y) / ymag + yc;
	    colour := (MaxSquares (z) * Skip MOD NumColours) + ColourOffset;
	    HiResDot (x, y, colour);
          END; (* WITH  *)
	END;
	INC (x,1);
      END; (* FOR *)
      done := EventHappened ();
      INC (y,1);
    END; (* FOR *)
  END DrawMandelbrot;

  PROCEDURE GetNewParams;

  VAR
    oldScale, x0, y0, x1, y1, button : INTEGER;

  BEGIN (* GetNewParams *)
    GetMouseClick (x0, y0, button);
    GetMouseClick (x1, y1, button);
 (* WriteInt (button, 1); WriteInt (button, 10); WriteLn; *)
    xc := FLOAT (x0 - xMax DIV 2) / xmag + xc;
    yc := FLOAT (yMax DIV 2 - y0) / ymag + yc;
    xmag := xmag * FLOAT (xMax) / FLOAT (2 * ABS (x1 - x0));
    ymag := xmag;
    IF xmag > 4000.0 THEN
      squaresLimit := 512;
    ELSIF xmag > 1000.0 THEN
      squaresLimit := 256;
    END;
    WriteString ("Centre: "); WriteReal (xc, 15);
    WriteString (", "); WriteReal (yc, 15);
    WriteString (";  magnification: "); WriteReal (xmag, 15);
    WriteLn;
  END GetNewParams;

  PROCEDURE InitParams;

  VAR
    numArgs : CARDINAL;
    argString : ARRAY [1..80] OF CHAR;

  BEGIN (* InitParams *)
    numArgs := ArgNumber();
    IF numArgs = 5 THEN
      (* Window size and location *)
      GetArg (1, argString);
      left := Value (argString);
      GetArg (2, argString);
      top := Value (argString);
      GetArg (3, argString);
      xMax := Value (argString);
      GetArg (4, argString);
      yMax := Value (argString);
    ELSIF numArgs = 1 THEN
      left := 600;
      top := 0;
      xMax := 400;
      yMax := 400;
    ELSE
      WriteString ("Usage: coarsemand [top left xSize ySize]");
      WriteLn;
      HALT;
    END;
    xc := 0.0;
    yc := 0.0;
    xmag := 100.0;
    ymag := xmag;
    magnitudeLimit := 4.0;
    squaresLimit := 100;
  END InitParams;

BEGIN (* CoarseMand *)
  InitParams;
  PlaceGraphWindow ("Mandelbrot", left, top, xMax, yMax);
  DrawBlock (300,300,200,200,30);
  REPEAT
    DrawMandelbrot;
    Flush;
    GetNewParams;
  UNTIL FALSE;
  CloseScreen;
END CoarseMand.
