MODULE Pento;

(*
 *  This program is copyright (c) Raoul Gough 1985 - 1990
 *
 *  Permission is granted to QUT to distribute this program
 *  as a demonstration program with the Gardens Point Modula
 *  compiler system
 *
 *  This program may be copied and or modified by users 
 *  for non-commercial purposes provided that this copyright 
 *  notice remains in the text
 *
 *  All other uses require the written permission of the 
 *  copyright holder.
 *)

FROM SYSTEM IMPORT SHIFT;
FROM InOut IMPORT Write, WriteLn, WriteString, WriteCard, ReadCard;
FROM Graphics IMPORT GraphWindow, DrawBlock, Flush;
FROM UxProcesses IMPORT Sleep;
FROM ProgArgs IMPORT ArgNumber, GetArg, Assert;

CONST
  numPieces = 12;
  graphBlockSize = 30;

TYPE
  Piece = ARRAY [0..4] OF BITSET;

  PieceOrientations = ARRAY [0..7] OF Piece;

  Pieces = RECORD
             numOrients : CARDINAL;
             orientations : PieceOrientations;
           END; (* Pieces *)

  Shifts = ARRAY [0..15] OF RECORD
             piece : ARRAY [0..11] OF Pieces;
           END; (* Shifts *)

CONST
  yMax = 13;  (*  6 + 4*2 - 1   (-1 : number from 0-) *)
  xMax = 18;  (* 10 + 4*2 - 1   (+4*2 for the border) *)
  width = (xMax + 15) DIV 16;

TYPE
  Board = ARRAY [0..yMax], [0..width] OF BITSET;

CONST
  FullBitSet = BITSET {0..31};
  LeftBorder = BITSET {12..31};
  RightBorder = BITSET {0, 1};
  EmptySet = BITSET {};

TYPE
  StorageRecord = ARRAY [0..11] OF RECORD
                    pNum, pOrient, xPos, yPos : CARDINAL;
                  END; (* StorageRecord *)

VAR
  display, noWindow, input : BOOLEAN;
  shift : Shifts;
  thisBoard : Board;
  numSols, numUsed : CARDINAL;
  solution : ARRAY [0..7] OF ARRAY [0..16] OF CHAR;
  solChars : ARRAY [0..11] OF CHAR;
  used : ARRAY [0..11] OF BOOLEAN;
  colour : ARRAY [0..11] OF CARDINAL;
  stored : StorageRecord;

  PROCEDURE Include (VAR board : Board; xBase, yBase:CARDINAL; piece:Piece);

  VAR
    x, y:CARDINAL;

  BEGIN (* Include *)
    x:=xBase DIV 16;
    FOR y:=0 TO 4 DO
      board [yBase,x]:=board [yBase,x] + piece[y];
      INC (yBase);
    END; (* FOR *)
  END Include;

  PROCEDURE Exclude (VAR board : Board; xBase, yBase:CARDINAL; piece:Piece);

  VAR
    x, y:CARDINAL;

  BEGIN (* Exclude *)
    x:=xBase DIV 16;
    FOR y:=0 TO 4 DO
      board [yBase,x]:=board [yBase,x] - piece[y];
      INC (yBase);
    END; (* FOR *)
  END Exclude;

  PROCEDURE StorePiece (xBase, yBase: CARDINAL; piece:Piece);

  BEGIN (* StorePiece *)
    Include (thisBoard, xBase, yBase, piece);
    INC (numUsed);
  END StorePiece;

  PROCEDURE RemovePiece (xBase, yBase: CARDINAL; piece:Piece);

  BEGIN (* RemovePiece *)
    Exclude (thisBoard, xBase, yBase, piece);
    DEC (numUsed);
  END RemovePiece;

  PROCEDURE GetNextHole ( VAR board : Board; (* INPUT only *)
                          startX, startY : CARDINAL;
                          VAR holeX, holeY : CARDINAL;
                          VAR found : BOOLEAN);

  BEGIN (* GetNextHole *)
    holeX:=startX;
    holeY:=startY;
    REPEAT
      IF holeY < yMax THEN holeY:=holeY + 1
      ELSIF holeX >= 15 THEN
        found := FALSE;
        RETURN;
      ELSE
        holeY:=0;
        holeX:=holeX+1;
      END; (* IF *)
    UNTIL NOT ((15 - holeX) IN board [holeY, 0]);
    found := TRUE;
  END GetNextHole;

  PROCEDURE PieceFills (x, y, xBase, yBase:CARDINAL; VAR piece:Piece) : BOOLEAN;

    (* RETURNS true iff piece may be placed at xBase, yBase AND fills
            x,y AND leavs no holes of size co-prime with 5 
            Note - starts search for dud holes from (x,y) UP and RIGHT *)

  VAR
    tempBoard : Board;
    setNum : CARDINAL;
    badHoles : BOOLEAN;

    PROCEDURE OverLap () : BOOLEAN;

    VAR
      yCount : CARDINAL;

    BEGIN (* OverLap *)
      FOR yCount:=0 TO 4 DO
        IF piece [yCount] * thisBoard [yBase + yCount, setNum] # BITSET {} THEN
          RETURN TRUE
        END; (* IF *)
      END; (* FOR *)
      RETURN FALSE;
    END OverLap;

    PROCEDURE CountHoles;

    VAR
      startX, startY, total : CARDINAL;
      found : BOOLEAN;

      PROCEDURE SearchHole (x, y : CARDINAL);

      VAR
        bitPos, setNum : CARDINAL;

      BEGIN (* SearchHole *)
        bitPos := 15 - x;
        setNum := 0;
        IF bitPos IN tempBoard [y, setNum] THEN
          (* Do Nothing *)
        ELSE
          INC (total);
          INCL (tempBoard [y, setNum], bitPos);
          SearchHole (x, y-1);
          SearchHole (x, y+1);
          SearchHole (x-1, y);
          SearchHole (x+1, y);
        END; (* SearchHole *)
      END SearchHole;

    BEGIN (* CountHoles *)
      startY:=y;
      startX:=x;
      badHoles:=FALSE;
      REPEAT
        GetNextHole (tempBoard, startX, startY, startX, startY, found);
        IF found THEN
          total := 0;
          SearchHole (startX, startY);
          badHoles := total MOD 5 > 0;
        END;
      UNTIL badHoles OR NOT found;
    END CountHoles;

  BEGIN (* PieceFills *)
    setNum := xBase DIV 16;
    IF OverLap () THEN
      RETURN FALSE;
    ELSE
      tempBoard := thisBoard;
      Include (tempBoard, xBase, yBase, piece);
      IF (15 - x) IN tempBoard [y, 0] THEN
        CountHoles;
        RETURN NOT badHoles;
      ELSE
        RETURN FALSE;
      END;
    END; (* IF *)
  END PieceFills;

  PROCEDURE StoreSolution ( xBase, yBase : CARDINAL;
                            piece : Piece;
                            num : CARDINAL);

    (* Store solChars [num] in the solution at elements of piece, placed
       at xBase, yBase *)

  VAR
    x,y : CARDINAL;
    c : CARDINAL;
    char : CHAR;

    PROCEDURE DrawSquare (x, y, colour : CARDINAL);

    BEGIN (* DrawSquare *)
      DrawBlock ((x - 15) * graphBlockSize, (y - 10) * graphBlockSize,
			graphBlockSize, graphBlockSize, colour);
    END DrawSquare;

  BEGIN (* StoreSolution *)
    IF num <> MAX(CARDINAL) THEN
      c := colour [num];
      char := solChars [num];
    ELSE
      c := 0;
      char := " ";
    END; (* IF *)
    FOR y:=0 TO 4 DO
      FOR x:=11 TO 0 BY -1 DO
        IF x IN piece [y] THEN
	  DrawSquare (27 - x, 20 - yBase - y, c);
	  solution [yBase + y - 3, 12 - x] := char;
        END; (* IF *)
      END; (* FOR *)
    END; (* FOR *)
    Flush;
  END StoreSolution;

  PROCEDURE OutputSolution;

  VAR
    y, count, n, throw : CARDINAL;

  BEGIN (* OutputSolution *)
    FOR count:=0 TO numUsed - 1 DO
      WITH stored [count] DO
        StoreSolution ( xPos, yPos,
                        shift [xPos MOD 16].piece[pNum].orientations[pOrient],
                        pNum);
      END; (* WITH *)
    END; (* FOR *)
(*
 *  n := numSols;
 *  FOR count:=0 TO 4 DO
 *    screen [24, 39 - count] := ORD ("0") + n MOD 10 + 256 * 7;
 *    n := n DIV 10;
 *  END; (* FOR *) 
 * 
 *  FOR y:=7 TO 0 BY -1 DO WriteString (solution[y]); WriteLn; END; WriteLn;
 *)
    IF display THEN throw := Sleep (5); END; (* Only pause when showing
  						continuous display *)
  END OutputSolution;

  PROCEDURE Fill (xBase, y : CARDINAL);

    (* Finds a piece which fills the (single square) hole AT y, xBase
       -- only enters piece if it leaves no Dud holes.  If sucess
       then recurses. *)

  VAR
    yBase, pieceNum, thisOrient, newX, newY : CARDINAL;
    thisPiece : Piece;
    found : BOOLEAN;

  BEGIN (* Fill *)
    WITH shift [xBase] DO
      FOR pieceNum:=0 TO 11 DO
        IF NOT used [pieceNum] THEN
          WITH piece [pieceNum] DO
            FOR thisOrient:=0 TO numOrients - 1 DO
              thisPiece:=orientations [thisOrient];
              FOR yBase:=y - 3 TO y DO
                IF PieceFills (xBase, y, xBase, yBase, thisPiece) THEN
                  WITH stored [numUsed] DO
                    pNum := pieceNum;
                    pOrient := thisOrient;
                    xPos := xBase;
                    yPos := yBase;
                  END; (* WITH *)
                  StorePiece (xBase, yBase, thisPiece);
                  IF display THEN
                    StoreSolution (xBase, yBase, thisPiece, pieceNum);
                  END; (* IF *)
                  used [pieceNum]:=TRUE;
                  IF numUsed = numPieces THEN
                    INC (numSols);
		   (*
		    * IF NOT display THEN WriteCard (numSols, 8); WriteLn END;
		    *)
                    IF NOT noWindow THEN OutputSolution END;
                  ELSE
                    GetNextHole (thisBoard, xBase, y, newX, newY, found);
		    Assert(found, "no hole found");
                    Fill (newX, newY);
                  END; (* IF *)
                  RemovePiece (xBase, yBase, thisPiece);
                  IF display THEN
                    StoreSolution (xBase, yBase, thisPiece, MAX(CARDINAL));
                  END; (* IF *)
                  used [pieceNum]:=FALSE;
                END; (* IF *)
              END; (* FOR *)
            END; (* FOR *)
          END; (* WITH *)
        END; (* IF *)
      END; (* FOR *)
    END; (* WITH *)
  END Fill;

  PROCEDURE InitBoard;

  VAR
    x, y, count : CARDINAL;

  BEGIN (* InitBoard *)
    FOR y:=4 TO yMax - 4 DO
      thisBoard [y, 0]:=LeftBorder + RightBorder;
      thisBoard [y, 1]:=FullBitSet;
    END; (* FOR *)
    FOR y:=0 TO 3 DO
      FOR x:=0 TO width DO
        thisBoard [y, x]:=FullBitSet;
        thisBoard [yMax - y, x]:=FullBitSet;
      END; (* FOR *)
    END; (* FOR *)
    solution [0]:="++++++++++++";
    FOR y:=1 TO 6 DO
      solution [y]:="+          +";
    END; (* FOR *)
    solution [7]:="++++++++++++";
    solChars := "1234567890!@";
    FOR count := 0 TO 11 DO used [count]:=FALSE; END;
    numUsed := 0;
    numSols := 0;
    colour[0]:=22;
    colour[1]:=32;
    colour[2]:=41;
    colour[3]:=51;
    colour[4]:=61;
    colour[5]:=71;
    colour[6]:=19;
    colour[7]:=29;
    colour[8]:=38;
    colour[9]:=48;
    colour[10]:=58;
    colour[11]:=68;
  END InitBoard;

  PROCEDURE DumpBoard(VAR this : Board);

  VAR
    x, y : CARDINAL;

  BEGIN (* DumpBoard *)
    FOR y:=0 TO yMax DO
      FOR x:=0 TO xMax DO
        IF (15 - x MOD 16) IN this[y, x DIV 16] THEN
          Write ("*")
        ELSE
          Write (" ");
        END; (* IF *)
      END; (* FOR *)
      WriteLn;
    END; (* FOR *)
    WriteLn;
  END DumpBoard;

  PROCEDURE DumpPiece(VAR this : Piece);
  VAR
    x, y : CARDINAL;
  BEGIN
    FOR y:=0 TO 3 DO
      FOR x := 0 TO 15 DO
        IF (15 - x) IN this[y] THEN
          Write ("*")
        ELSE
          Write (" ");
        END; (* IF *)
      END; (* FOR *)
      WriteLn;
    END; (* FOR *)
    WriteLn;
  END DumpPiece;

  PROCEDURE FindSolutions;

  VAR
    thisOrient, x1, y1, x, y, xMax, yMax : CARDINAL;
    found : BOOLEAN;
    thisPiece : Piece;

  BEGIN (* FindSolutions *)
    xMax := 6;
    yMax := 6;
    used [0] := TRUE;
    WITH stored [0] DO
      pNum := 0;
      FOR thisOrient := 0 TO 1 DO
        pOrient := thisOrient;
        FOR y1 := 4 TO yMax DO
          yPos := y1;
          FOR x1 := 4 TO xMax DO
            xPos := x1;
            thisPiece := shift[xPos].piece[0].orientations[pOrient];
            StorePiece ( xPos, yPos, thisPiece);
            IF display THEN
              StoreSolution ( xPos, yPos, thisPiece, 0);
            END; (* IF *)
            GetNextHole (thisBoard, 0, 0, x, y, found);
            Fill (x, y);
            RemovePiece (xPos, yPos, thisPiece);
            IF display THEN
              StoreSolution (xPos, yPos, thisPiece, MAX (CARDINAL));
            END; (* IF *)
          END; (* FOR *)
        END; (* FOR *)
        yMax := 4;
        xMax := 8;
      END; (* FOR *)
    END; (* WITH *)
  END FindSolutions;

  PROCEDURE InitPieces;

  VAR
    pieceNum, count, o, y, ocount, setcount : CARDINAL;

    PROCEDURE Flip (VAR piece : Pieces; inOrient, outOrient, max : CARDINAL);

    VAR
      count : CARDINAL;

    BEGIN (* Flip *)
      WITH piece DO
        FOR count:=0 TO max DO
          orientations [outOrient][count]:=orientations [inOrient][max-count];
        END; (* FOR *)
      END; (* WITH *)
    END Flip;

  BEGIN (* InitPieces *)
    WITH shift [0] DO

      FOR pieceNum := 0 TO 11 DO
        WITH piece [pieceNum] DO
          FOR o := 0 TO 7 DO
            FOR y := 0 TO 4 DO
              orientations [o][y] := EmptySet; (* Init all to empty *)
            END; (* FOR *)
          END; (* FOR *)
        END; (* WITH *)
      END; (* FOR *)

      WITH piece [0] DO
        numOrients:=2;
        orientations [0][0]:=BITSET {15,14,13,12,11};
        FOR y:=0 TO 4 DO orientations [1][y]:=BITSET {15}; END;
      END; (* WITH piece [0] *)

      WITH piece [1] DO
        numOrients:=4;
        orientations [0][0] := BITSET {15        };
        orientations [0][1] := BITSET {15, 14    };
        orientations [0][2] := BITSET {    14, 13};

        orientations [1][0] := BITSET {        13};
        orientations [1][1] := BITSET {    14, 13};
        orientations [1][2] := BITSET {15, 14    };
        Flip (piece[1], 0, 2, 2);
        Flip (piece[1], 1, 3, 2);
      END; (* WITH piece [1] *)

      WITH piece [2] DO
        numOrients := 4;
        orientations [0][0] := BITSET {15        };
        orientations [0][1] := BITSET {15        };
        orientations [0][2] := BITSET {15, 14, 13};

        orientations [1][0] := BITSET {        13};
        orientations [1][1] := BITSET {        13};
        orientations [1][2] := BITSET {15, 14, 13};
        Flip (piece[2], 0, 2, 2);
        Flip (piece[2], 1, 3, 2);
      END; (* WITH piece [2] *)

      WITH piece [3] DO
        numOrients := 1;
        orientations [0][0] := BITSET {    14    };
        orientations [0][1] := BITSET {15, 14, 13};
        orientations [0][2] := BITSET {    14    };
      END; (* WITH piece [3] *)

      WITH piece [4] DO
        numOrients := 4;
        orientations [0][0] := BITSET {15, 14, 13};
        orientations [0][1] := BITSET {    14    };
        orientations [0][2] := BITSET {    14    };

        Flip (piece[4], 0, 1, 2);

        orientations [2][0] := BITSET {        13};
        orientations [2][1] := BITSET {15, 14, 13};
        orientations [2][2] := BITSET {        13};

        orientations [3][0] := BITSET {15        };
        orientations [3][1] := BITSET {15, 14, 13};
        orientations [3][2] := BITSET {15        };
      END; (* WITH piece [4] *)

      WITH piece [5] DO
        numOrients := 8;
        orientations [0][0] := BITSET {        13, 12};
        orientations [0][1] := BITSET {15, 14, 13    };
        Flip (piece[5], 0, 1, 1);

        orientations [2][0] := BITSET {15, 14        };
        orientations [2][1] := BITSET {    14, 13, 12};
        Flip (piece[5], 2, 3, 1);

        orientations [4][0] := BITSET {15    };
        orientations [4][1] := BITSET {15    };
        orientations [4][2] := BITSET {15, 14};
        orientations [4][3] := BITSET {    14};
        Flip (piece[5], 4, 5, 3);

        orientations [6][0] := BITSET {    14};
        orientations [6][1] := BITSET {    14};
        orientations [6][2] := BITSET {15, 14};
        orientations [6][3] := BITSET {15    };
        Flip (piece[5], 6, 7, 3);
      END; (* WITH piece [5] *)

      WITH piece [6] DO
        numOrients := 8;
        orientations [0][0] := BITSET {    14    };
        orientations [0][1] := BITSET {15, 14, 13};
        orientations [0][2] := BITSET {15        };
        Flip (piece[6], 0, 1, 2);

        orientations [2][0] := BITSET {    14    };
        orientations [2][1] := BITSET {15, 14, 13};
        orientations [2][2] := BITSET {        13};
        Flip (piece[6], 2, 3, 2);

        orientations [4][0] := BITSET {15, 14    };
        orientations [4][1] := BITSET {    14, 13};
        orientations [4][2] := BITSET {    14    };
        Flip (piece[6], 4, 5, 2);

        orientations [6][0] := BITSET {    14, 13};
        orientations [6][1] := BITSET {15, 14    };
        orientations [6][2] := BITSET {    14    };
        Flip (piece[6], 6, 7, 2);
      END; (* WITH piece [6] *)

      WITH piece [7] DO
        numOrients := 4;
        orientations [0][0] := BITSET {        13};
        orientations [0][1] := BITSET {15, 14, 13};
        orientations [0][2] := BITSET {15        };
        Flip (piece[7], 0, 1, 2);

        orientations [2][0] := BITSET {15, 14    };
        orientations [2][1] := BITSET {    14    };
        orientations [2][2] := BITSET {    14, 13};
        Flip (piece[7], 2, 3, 2);
      END; (* WITH piece [7] *)

      WITH piece [8] DO
        numOrients := 4;
        orientations [0][0] := BITSET {15, 14};
        orientations [0][1] := BITSET {    14};
        orientations [0][2] := BITSET {15, 14};

        orientations [1][0] := BITSET {15, 14};
        orientations [1][1] := BITSET {15    };
        orientations [1][2] := BITSET {15, 14};

        orientations [2][0] := BITSET {15, 14, 13};
        orientations [2][1] := BITSET {15,     13};
        Flip (piece[8], 2, 3, 1);
      END; (* WITH piece [8] *)

      WITH piece [9] DO
        numOrients := 8;
        orientations [0][0] := BITSET {15, 14, 13, 12};
        orientations [0][1] := BITSET {    14        };
        Flip (piece[9], 0, 1, 1);

        orientations [2][0] := BITSET {15, 14, 13, 12};
        orientations [2][1] := BITSET {        13    };
        Flip (piece[9], 2, 3, 1);

        orientations [4][0] := BITSET {15    };
        orientations [4][1] := BITSET {15    };
        orientations [4][2] := BITSET {15, 14};
        orientations [4][3] := BITSET {15    };
        Flip (piece[9], 4, 5, 3);

        orientations [6][0] := BITSET {    14};
        orientations [6][1] := BITSET {    14};
        orientations [6][2] := BITSET {15, 14};
        orientations [6][3] := BITSET {    14};
        Flip (piece[9], 6, 7, 3);
      END; (* WITH piece [9] *)

      WITH piece [10] DO
        numOrients := 8;
        orientations [0][0] := BITSET {15, 14, 13, 12};
        orientations [0][1] := BITSET {15            };
        Flip (piece[10], 0, 1, 1);

        orientations [2][0] := BITSET {15, 14, 13, 12};
        orientations [2][1] := BITSET {            12};
        Flip (piece[10], 2, 3, 1);

        orientations [4][0] := BITSET {15    };
        orientations [4][1] := BITSET {15    };
        orientations [4][2] := BITSET {15    };
        orientations [4][3] := BITSET {15, 14};
        Flip (piece[10], 4, 5, 3);

        orientations [6][0] := BITSET {    14};
        orientations [6][1] := BITSET {    14};
        orientations [6][2] := BITSET {    14};
        orientations [6][3] := BITSET {15, 14};
        Flip (piece[10], 6, 7, 3);
      END; (* WITH piece [10] *)

      WITH piece [11] DO
        numOrients := 8;
        orientations [0][0] := BITSET {15, 14, 13};
        orientations [0][1] := BITSET {15, 14    };
        Flip (piece[11], 0, 1, 1);

        orientations [2][0] := BITSET {15, 14, 13};
        orientations [2][1] := BITSET {    14, 13};
        Flip (piece[11], 2, 3, 1);

        orientations [4][0] := BITSET {15, 14};
        orientations [4][1] := BITSET {15, 14};
        orientations [4][2] := BITSET {15    };
        Flip (piece[11], 4, 5, 2);

        orientations [6][0] := BITSET {15, 14};
        orientations [6][1] := BITSET {15, 14};
        orientations [6][2] := BITSET {    14};
        Flip (piece[11], 6, 7, 2);
      END; (* WITH piece [11] *)
    END; (* WITH *)

    FOR count:=1 TO 15 DO
      WITH shift [count] DO
        piece:=shift [count-1].piece;
        FOR pieceNum:=0 TO 11 DO
 	  WITH piece[pieceNum] DO
	    FOR ocount := 0 TO numOrients - 1 DO
	      FOR setcount := 0 TO 4 DO
		SHIFT (orientations [ocount, setcount], -1);
	      END; (* FOR *)
	    END; (* FOR *)
	  END; (* WITH *)
        END; (* FOR *)
      END; (* WITH *)
    END; (* FOR *)
  END InitPieces;

  PROCEDURE GetDisplayState;

  VAR
    str : ARRAY [0..5] OF CHAR;

  BEGIN (* GetDisplayState *)
    IF ArgNumber() = 1 THEN
      display  := FALSE;
      noWindow := FALSE;
      WriteString ('Use "pento -d" for continuous display'); WriteLn;
      WriteString ('Use "pento -n" for just a solution count'); WriteLn;
    ELSE
      GetArg (1, str);
      display  := (str[0] = "-") AND ((str[1] = "d") OR (str[1] = "D"));
      noWindow := (str[0] = "-") AND ((str[1] = "n") OR (str[1] = "N"));
    END; (* IF *)
  END GetDisplayState;

BEGIN (* Pento *)
  input:=FALSE;
  GetDisplayState;
  IF NOT noWindow THEN
    GraphWindow("pento", graphBlockSize * 12, graphBlockSize * 8);
  END;
  InitPieces;
  InitBoard;
  FindSolutions;
  IF noWindow THEN
    WriteString("Number of distinct solutions = ");
    WriteCard(numSols,0);
    WriteLn;
  END;
END Pento.
